# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Main application entry point."""

# yapf: disable

# Standard library imports
import os


# yapf: enable

PATH_APP = os.path.dirname(os.path.abspath(__file__))
